/* 
 * item_interpose.c -- demonstration of interposition on Panel_items.
 * Richard M. Goldstein, Window Developer Environment
 * SunSoft, Inc.
 * April 1992
 */


/*
 * Example of a generic method of interposing on Panel_item's
 * without modifying the XView library to support this directly.
 */

#include <stdio.h>
#include <xview/xview.h>
#include <xview/panel.h>
#include <xview/notify.h>


static Notify_value 	base_item_event_handler();
static Notify_value 	decimal_numbers_only();
static Notify_value 	no_periods();
static Notify_value 	no_lower_case();
static void		post_panel_item_events();
static void		register_panel_items();


static Attr_attribute EVENT_PROC_KEY;


#define isdot(chr)	((chr) == '.')


/********************************************************************************/


void
main ( argc,argv )
     int argc;
     char **argv;
{
    Frame frame;
    Panel panel;
    Panel_text_item upper_text;
    Panel_text_item number_text;
    Panel_text_item decimal_text;

    (void) xv_init ( XV_INIT_ARGC_PTR_ARGV, &argc, argv, NULL );
    EVENT_PROC_KEY = xv_unique_key();

    frame = xv_create ( XV_NULL, FRAME,
		       XV_LABEL,	"Modified Textfields",
		       NULL );

    panel = xv_create ( frame, PANEL, 
		       PANEL_LAYOUT,	PANEL_VERTICAL,
		       NULL );

    upper_text = xv_create ( panel, PANEL_TEXT, 
			    PANEL_LABEL_STRING,		"Upper Case:",
			    PANEL_VALUE_DISPLAY_LENGTH,	16,
			    NULL );

    decimal_text = xv_create ( panel, PANEL_TEXT, 
			      PANEL_LABEL_STRING,		"Decimal:",
			      PANEL_VALUE_DISPLAY_LENGTH,	12,
			      NULL );

    number_text = xv_create ( panel, PANEL_TEXT, 
			     PANEL_LABEL_STRING,		"Integer:",
			     PANEL_VALUE_DISPLAY_LENGTH,	12,
			     NULL );


    /* register panel item's with the notifier */
    register_panel_items ( panel );


    /* application specific handlers */
    notify_interpose_event_func ( upper_text, no_lower_case, NOTIFY_SAFE );
    notify_interpose_event_func ( decimal_text, decimal_numbers_only, NOTIFY_SAFE );

    /* 
     * implement integer by using existing decimal handler and weeding
     * out the periods.  note:  interposers are stacked, so order is
     * important.
     */
    notify_interpose_event_func ( number_text, decimal_numbers_only, NOTIFY_SAFE );
    notify_interpose_event_func ( number_text, no_periods, NOTIFY_SAFE );

    window_fit ( panel );
    window_fit ( frame );
    xv_main_loop ( frame );
    exit ( 0 );
}


/********************************************************************************/
/* These routines are generic to the panel items.				*/
/********************************************************************************/


/*
 * Traverse each item in the given Panel and register our base event handler
 * for it with the Notifier.  also, replace the default event handler with
 * one that posts events via the Notifier and attch the old one onto the
 * item with key-data.  Note:  once this is done, our Panel_item's are
 * functionally no different than when we started, except that they can
 * now be interposed on.
 */
static void
register_panel_items ( panel )
     Panel panel;
{
    Panel_item item;
    void (* items_event_proc)();

    PANEL_EACH_ITEM(panel, item)
	items_event_proc = (void (*)()) xv_get ( item, PANEL_EVENT_PROC );
        xv_set ( item, 
		PANEL_EVENT_PROC,	post_panel_item_events,
		XV_KEY_DATA, 	EVENT_PROC_KEY, items_event_proc, 
		NULL );
        notify_set_event_func ( item, base_item_event_handler, NOTIFY_SAFE );
    PANEL_END_EACH

}



/*
 * this routine is "generic" in that it will work
 * with any panel item, presuming you set the equally 
 * generic "base_item_event_handler" on each panel item 
 * you use this with.
 */
static void
post_panel_item_events ( item, event )
     Panel_item item;
     Event *event;
{
    /* 
     * rather than let the brain-damaged Panel delegate
     * the event directly, post it to the panel item so 
     * that an interposition stream can deal with it.
     */
    notify_post_event ( (Notify_client) item, 
		       (Notify_event) event, 
		       NOTIFY_SAFE );
}



/*
 * The generic base handler for panel items.  this makes
 * a direct call to the items default event handler, just
 * as the Panel would normally do.  Note:  a base handler
 * is always the LAST handler to see the event.
 */
static Notify_value 
base_item_event_handler ( client, event, arg, type )
     Notify_client client;
     Notify_event event;
     Notify_arg arg;
     Notify_event_type type; 
{
    void (* items_event_proc)() 
	= (void (*)())xv_get ( (Panel_item) client, XV_KEY_DATA, EVENT_PROC_KEY );

    if ( items_event_proc )
	(* items_event_proc)( (Panel_item) client, (Event *)event );

    /*
     * Assume the panel item handled the event, as the Panel 
     * doesn't return any useful status information.
     */
    return NOTIFY_DONE;
}



/********************************************************************************/
/* This routines are for item/application specific purposes.			*/
/* For the sake of example, they don't attempt to handle all possble events,	*/
/* rather, they are intended to be simple and understandable.			*/
/********************************************************************************/


/*
 * An interposer that can weed out all printable characters
 * that don't comprise decimal values for a Textfield.
 */
static Notify_value 
decimal_numbers_only ( client, event, arg, type )
     Notify_client client;
     Notify_event event;
     Notify_arg arg;
     Notify_event_type type; 
{
    int id = event_action ( (Event *)event );
    int found_decimal = FALSE;
    char *value;
    char *ptr;

    /*
     * don't bother to muck with editing events, just
     * make sure they get through.
     */
    if ( !isprint(id) )
	return notify_next_event_func ( client, event, arg, type );

    /*
     * For good measure, make sure there is not
     * already a decimal point in the item.
     */
    value = (char *) xv_get ( (Panel_item)client, PANEL_VALUE );
    for ( ptr=value; *ptr; ptr++ )
	if ( isdot(*ptr) ) {
	    found_decimal = TRUE;
	    break;
	}

    if ( isdigit(id) || (isdot(id) && !found_decimal) )
	return notify_next_event_func ( client, event, arg, type );

    /* Invalid Character! */
    if ( event_is_down((Event *)event) )
	window_bell( (Panel) xv_get((Panel_item) client, XV_OWNER) );
    return NOTIFY_DONE;
}




/*
 * An interposer that can filter out the '.' characters
 */
static Notify_value 
no_periods ( client, event, arg, type )
     Notify_client client;
     Notify_event event;
     Notify_arg arg;
     Notify_event_type type; 
{
    int id = event_action ( (Event *)event );

    if ( !isdot(id) )
	return notify_next_event_func ( client, event, arg, type );

    /* No '.' allowed! */
    if ( event_is_down((Event *)event) )
	window_bell( (Panel) xv_get((Panel_item) client, XV_OWNER) );
    return NOTIFY_DONE;
}



/*
 * Force all characters to be upper case
 */
static Notify_value 
no_lower_case ( client, event, arg, type )
     Notify_client client;
     Notify_event event;
     Notify_arg arg;
     Notify_event_type type; 
{
    int id = event_action ( (Event *)event );

    /*
     * it is perfectly legal to modify an event from
     * an interposer.
     */
    if ( !(isalpha(id) && isupper(id)) )
	event_set_action ( (Event *)event, toupper(id) );

    return notify_next_event_func ( client, event, arg, type );
}


